@@ -22,18 +22,19 @@ from difflib import unified_diff
 import tempfile
 import ast
 
+from distutils.version import StrictVersion
 try:
     import pep8
-    if not pep8.__version__ >= '0.5.1':
+    if StrictVersion(pep8.__version__) < StrictVersion('0.5.1'):
         pep8 = None
 except ImportError:
     pep8 = None
 
 
-__version__ = '0.6'
+__version__ = '0.6.1'
 
 
-pep8bin = 'pep8'
+PEP8_BIN = 'pep8'
 PEP8_PASSES_MAX = 100
 CR = '\r'
 LF = '\n'
@@ -88,35 +89,39 @@ class FixPEP8(object):
         self.fix_e223 = self.fix_e221
         self.fix_e241 = self.fix_e221
         self.fix_e242 = self.fix_e224
+        self.fix_e261 = self.fix_e262
         self.fix_w191 = self.fix_e101
 
     def _spawn_pep8(self, targetfile):
         """execute pep8 via subprocess.Popen."""
         paths = os.environ['PATH'].split(':')
         for path in paths:
-            if os.path.exists(os.path.join(path, pep8bin)):
-                cmd = ([os.path.join(path, pep8bin)] +
+            if os.path.exists(os.path.join(path, PEP8_BIN)):
+                cmd = ([os.path.join(path, PEP8_BIN)] +
                        self._pep8_options(targetfile))
                 p = Popen(cmd, stdout=PIPE)
-                return p.stdout.readlines()
-        raise Exception("'%s' is not found." % pep8bin)
+                return [l.decode('utf8') for l in p.stdout.readlines()]
+        raise Exception("'%s' is not found." % PEP8_BIN)
 
     def _execute_pep8(self, targetfile):
         """execute pep8 via python method calls."""
         pep8.options, pep8.args = \
                 pep8.process_options(['pep8'] + self._pep8_options(targetfile))
+        # Override sys.stdout to get results from pep8
         sys_stdout = sys.stdout
-        fake_stdout = StringIO()
-        sys.stdout = fake_stdout
-        tmp_checker = pep8.Checker(self.filename, lines=self.source)
-        tmp_checker.check_all()
-        sys.stdout = sys_stdout
+        try:
+            fake_stdout = StringIO()
+            sys.stdout = fake_stdout
+            tmp_checker = pep8.Checker(self.filename, lines=self.source)
+            tmp_checker.check_all()
+        finally:
+            sys.stdout = sys_stdout
         result = fake_stdout.getvalue()
         return StringIO(result).readlines()
 
     def _pep8_options(self, targetfile):
         """return options to be passed to pep8."""
-        return (["-r", targetfile] +
+        return (["--repeat", targetfile] +
                 (["--ignore=" + self.options.ignore]
                  if self.options.ignore else []))
 
@@ -131,11 +136,12 @@ class FixPEP8(object):
                 modified_lines = fix(result)
                 if modified_lines:
                     completed_lines += modified_lines
-                if self.options.verbose:
-                    if modified_lines == []:
+                elif modified_lines == []:  # Empty list means no fix
+                    if self.options.verbose:
                         sys.stderr.write('Not fixing {f} on line {l}\n'.format(
                                 f=result['id'], l=result['line']))
-                completed_lines.append(result['line'])
+                else:  # We assume one-line fix when None
+                    completed_lines.append(result['line'])
             else:
                 if self.options.verbose:
                     sys.stderr.write("'%s' is not defined.\n" % \
@@ -164,8 +170,9 @@ class FixPEP8(object):
         """Reindent all lines."""
         reindenter = Reindenter(self.source)
         if reindenter.run():
+            original_length = len(self.source)
             self.source = reindenter.fixed_lines()
-            return range(len(self.source))
+            return range(1, 1 + original_length)
         else:
             return []
 
@@ -229,33 +236,35 @@ class FixPEP8(object):
         self.source[result['line'] - 1] = fixed
 
     def fix_e251(self, result):
-        target = self.source[result['line'] - 1]
+        line_index = result['line'] - 1
+        target = self.source[line_index]
         c = result['column'] - 1
         fixed = target[:c] + re.sub(r'\s*=\s*', '=', target[c:], 1)
-        self.source[result['line'] - 1] = fixed
 
-    def fix_e261(self, result):
-        target = self.source[result['line'] - 1]
-        c = result['column']
+        # There could be an escaped newline
+        #
+        #     def foo(a=\
+        #             1)
+        if (fixed.endswith('=\\\n') or
+            fixed.endswith('=\\\r\n') or
+            fixed.endswith('=\\\r')):
+            self.source[line_index] = fixed.rstrip('\n\r \t\\')
+            self.source[line_index + 1] = \
+                    self.source[line_index + 1].lstrip()
+            return [line_index + 1, line_index + 2]  # Line indexed at 1
 
-        # pep8 is sometimes off by one in cases like "{# comment"
-        if target[c] == '#':
-            pass
-        elif target[c - 1] == '#':
-            c = c - 1
-        else:
-            return
-
-        fixed = target[:c] + " " + target[c:]
         self.source[result['line'] - 1] = fixed
 
     def fix_e262(self, result):
+        """Fix spacing after comment hash."""
         target = self.source[result['line'] - 1]
-        split = target.rsplit('#', 1)
+        offset = result['column']
 
-        assert len(split) == 2
-        comment = split[1].lstrip()
-        fixed = split[0].rstrip(' \t#') + ('  # ' + comment if comment else '')
+        code = target[:offset].rstrip(' \t#')
+        comment = target[offset:].lstrip(' \t#')
+
+        fixed = code + ('  # ' + comment if comment.strip()
+                        else self.newline)
 
         self.source[result['line'] - 1] = fixed
 
@@ -287,7 +296,7 @@ class FixPEP8(object):
                 break
             if not self.source[line].strip():
                 self.source[line] = ''
-                modified_lines.append(line)
+                modified_lines.append(1 + line)  # Line indexed at 1
                 cnt += 1
             line -= 1
 
@@ -297,16 +306,11 @@ class FixPEP8(object):
         line_index = result['line'] - 1
         target = self.source[line_index]
 
-        # Take care of semicolons first
-        if target.rstrip().endswith(';'):
-            self.source[line_index] = _fix_multiple_statements(target,
-                                                               self.newline)
-        else:
-            indentation = target.split("import ")[0]
-            modules = target.split("import ")[1].split(",")
-            fixed_modulelist = \
-                    [indentation + "import %s" % m.lstrip() for m in modules]
-            self.source[line_index] = self.newline.join(fixed_modulelist)
+        indentation = target.split("import ")[0]
+        modules = target.split("import ")[1].split(",")
+        fixed_modulelist = \
+                [indentation + "import %s" % m.lstrip() for m in modules]
+        self.source[line_index] = self.newline.join(fixed_modulelist)
 
     def fix_e701(self, result):
         line_index = result['line'] - 1
@@ -319,15 +323,28 @@ class FixPEP8(object):
         self.source[result['line'] - 1] = fixed_source
 
     def fix_e702(self, result):
+        """Fix multiple statements on one line."""
         target = self.source[result['line'] - 1]
+
+        # We currently do not support things like
+        #     """
+        #         hello
+        #       """; foo()
         if '"""' in target:
-            # We currently do not support things like
-            #     """
-            #         hello
-            #       """; foo()
             return []
-        self.source[result['line'] - 1] = _fix_multiple_statements(target,
-                self.newline)
+
+        # Make sure we aren't likely in a string
+        if target.strip().startswith('"') or target.strip().startswith("'"):
+            return []
+
+        offset = result['column'] - 1
+        first = target[:offset].rstrip(';')
+        second = target[offset:].lstrip(';')
+
+        f = [_get_indentation(target) + t.strip()
+             for t in [first, second] if t.strip()]
+
+        self.source[result['line'] - 1] = self.newline.join(f) + self.newline
 
     def fix_w291(self, result):
         fixed_line = self.source[result['line'] - 1].rstrip()
@@ -343,19 +360,19 @@ class FixPEP8(object):
     def fix_w391(self, _):
         source = copy.copy(self.source)
         source.reverse()
-        found_notblank = False
         blank_count = 0
         for line in source:
             line = line.rstrip()
-            if re.match("^$", line):
-                blank_count += 1
-            else:
-                found_notblank = True
-            if found_notblank and not re.match("^$", line):
+            if line:
                 break
+            else:
+                blank_count += 1
         source = source[blank_count:]
         source.reverse()
+
+        original_length = len(self.source)
         self.source = source
+        return range(1, 1 + original_length)
 
     def fix_w601(self, result):
         target = self.source[result['line'] - 1]
@@ -365,7 +382,11 @@ class FixPEP8(object):
 
         # Skip complicated cases
         if target.count('(') > 1 or target.count(')') > 1:
-            return
+            return []
+        if target.count('(') != target.count(')'):
+            return []
+        if target.count(',') > 0:
+            return []
 
         _tmp = target.split(".has_key", 1)
 
@@ -401,6 +422,7 @@ class FixPEP8(object):
             "".join([_before, _arg, " in ", _symbol, _after])
 
     def fix_w602(self, result):
+        """Fix deprecated form of raising exception."""
         line_index = result['line'] - 1
         line = self.source[line_index]
 
@@ -409,27 +431,19 @@ class FixPEP8(object):
             # Give up
             return []
 
-        if line.rstrip().endswith(';'):
-            # Take care of semicolons first
-            self.source[line_index] = _fix_multiple_statements(line,
-                                                               self.newline)
-            return
-        elif line[-2:] == '\\\n':
-            # Remove escaped LF first
-            self.source[line_index] = line[:-2]
-            return
-        elif line[-3:] == '\\\r\n':
-            # Remove escaped CRLF first
-            self.source[line_index] = line[:-3]
-            return
-        elif line[-2:] == '\\\r':
-            # Remove escaped CR first
-            # NOTE: Doesn't get executed because pep8 doesn't seem to work with
-            #       CR line endings
-            self.source[line_index] = line[:-2]
-            return
+        if ' or ' in line or ' and ' in line:
+            # Give up
+            return []
 
-        modified_lines = [line_index]
+        if (line.endswith('\\\n') or
+            line.endswith('\\\r\n') or
+            line.endswith('\\\r')):
+            self.source[line_index] = line.rstrip('\n\r \t\\')
+            self.source[line_index + 1] = \
+                    ' ' + self.source[line_index + 1].lstrip()
+            return [line_index + 1, line_index + 2]  # Line indexed at 1
+
+        modified_lines = [1 + line_index]  # Line indexed at 1
 
         double = '"""'
         single = "'''"
@@ -461,7 +475,7 @@ class FixPEP8(object):
                 line_contents = self.source[i]
                 self.source[line_index] += line_contents
                 self.source[i] = ''
-                modified_lines.append(i)
+                modified_lines.append(1 + i)  # Line indexed at 1
             line = self.source[line_index]
 
         indent, rest = _split_indentation(line)
@@ -511,43 +525,8 @@ class FixPEP8(object):
                 return modified_lines
             else:
                 return []
-
-        sio = StringIO(line)
-        is_found_raise = False
-        first_comma_found = False
-        comment = ''
-        args = ''
-        indentation = ''
-        exception_type = None
-        for tokens in tokenize.generate_tokens(sio.readline):
-            if tokens[0] is token.INDENT:
-                assert not indentation
-                indentation = tokens[1]
-            elif tokens[1] == 'raise':
-                is_found_raise = True
-            elif tokens[0] is token.NAME and is_found_raise:
-                if exception_type:
-                    args += tokens[1]
-                else:
-                    exception_type = tokens[1]
-            elif tokens[0] is token.NEWLINE:
-                break
-            elif tokens[0] is not token.DEDENT:
-                if tokens[1].startswith(',') and not first_comma_found:
-                    first_comma_found = True
-                elif tokens[1].startswith('#'):
-                    assert not comment
-                    comment = tokens[1]
-                    break
-                else:
-                    args += tokens[1]
-        assert exception_type
-        self.source[result['line'] - 1] = \
-            ''.join([indentation, 'raise ', exception_type,
-                     '(',
-                     args[1:-1] if args.startswith('(') else args,
-                     ')',
-                     comment, self.newline])
+        else:
+            self.source[line_index] = _fix_basic_raise(line, self.newline)
 
         return modified_lines
 
@@ -613,12 +592,6 @@ def _split_indentation(line):
     return (line[:non_whitespace_index], line[non_whitespace_index:])
 
 
-def _fix_multiple_statements(target, newline):
-    f = [_get_indentation(target) + t.strip()
-         for t in target.split(";") if t.strip()]
-    return newline.join(f) + newline
-
-
 def _analyze_pep8result(result):
     tmp = result.split(":")
     filename = tmp[0]
@@ -641,11 +614,57 @@ def _priority_key(pep8_result):
     Global fixes should be done first. This is important for things
     like indentation.
     """
-    high_priority = ['e101', 'e111', 'w191']
-    return pep8_result['id'].lower() not in high_priority
+    priority = ['e101', 'e111', 'w191',  # Global fixes
+                'e702',  # Break multiline statements early
+               ]
+    key = pep8_result['id'].lower()
+    if key in priority:
+        return priority.index(key)
+    else:
+        # Lowest priority
+        return len(priority)
 
 
-class Reindenter:
+def _fix_basic_raise(line, newline):
+    """Fix W602 basic case"""
+    sio = StringIO(line)
+    is_found_raise = False
+    first_comma_found = False
+    comment = ''
+    args = ''
+    indentation = ''
+    exception_type = None
+    for tokens in tokenize.generate_tokens(sio.readline):
+        if tokens[0] is token.INDENT:
+            assert not indentation
+            indentation = tokens[1]
+        elif tokens[1] == 'raise':
+            is_found_raise = True
+        elif tokens[0] is token.NAME and is_found_raise:
+            if exception_type:
+                args += tokens[1]
+            else:
+                exception_type = tokens[1]
+        elif tokens[0] is token.NEWLINE:
+            break
+        elif tokens[0] is not token.DEDENT:
+            if tokens[1].startswith(',') and not first_comma_found:
+                first_comma_found = True
+            elif tokens[1].startswith('#'):
+                assert not comment
+                comment = tokens[1]
+                break
+            else:
+                args += tokens[1]
+    assert exception_type
+    return ''.join([indentation, 'raise ', exception_type,
+                    '(',
+                    args[1:-1] if args.startswith('(') else args,
+                    ')',
+                    comment, newline])
+
+
+class Reindenter(object):
     """Reindents badly-indented code to uniformly use four-space indentation.
     Released to the public domain, by Tim Peters, 03 October 2000."""
     def __init__(self, input_text):
@@ -838,12 +857,15 @@ def fix_file(filename, opts):
 
 def main():
     """tool main"""
-    parser = OptionParser(version="autopep8: %s" % __version__,
-                          description=__doc__)
+    parser = OptionParser(usage='Usage: autopep8 [options] '
+                                '[filename [filename ...]]',
+                          version="autopep8: %s" % __version__,
+                          description=__doc__,
+                          prog='autopep8')
     parser.add_option('-v', '--verbose', action='store_true', dest='verbose',
-                      help='print to verbose result.')
+                      help='print verbose messages')
     parser.add_option('-d', '--diff', action='store_true', dest='diff',
-                      help='diff print of fixed source.')
+                      help='print the diff for the fixed source')
     parser.add_option('-i', '--in-place', action='store_true',
                       help='make changes to files in place')
     parser.add_option('-p', '--pep8-passes',
@@ -854,14 +876,20 @@ def main():
                       help='do not fix these errors/warnings (e.g. E4,W)')
     opts, args = parser.parse_args()
     if not len(args):
-        print(parser.format_help())
-        return 1
+        parser.error('incorrect number of arguments')
 
-    if opts.in_place:
-        for f in args:
-            fix_file(f, opts)
-    else:
-        fix_file(args[0], opts)
+    try:
+        if opts.in_place:
+            for f in args:
+                fix_file(f, opts)
+        else:
+            if len(args) > 1:
+                parser.error('autopep8 only takes one filename as argument '
+                             'unless the "--in-place" option is used')
+            fix_file(args[0], opts)
+    except IOError as error:
+        sys.stderr.write(str(error) + '\n')
+        return 1
 
 
 if __name__ == '__main__':
