@@ -4,10 +4,20 @@ import unittest
 from subprocess import Popen, PIPE
 from tempfile import mkstemp
 
-sys.path.insert(0,
-                os.path.split(os.path.abspath(os.path.dirname(__file__)))[0])
+ROOT_DIR = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
+
+sys.path.insert(0, ROOT_DIR)
 import autopep8
 
+if 'AUTOPEP8_COVERAGE' in os.environ and int(os.environ['AUTOPEP8_COVERAGE']):
+    AUTOPEP8_CMD_TUPLE = ('coverage', 'run', '--branch', '--parallel',
+                          os.path.join(ROOT_DIR, 'autopep8.py'),)
+else:
+    # We need to specify the executable to make sure the correct Python
+    # interpreter gets used.
+    AUTOPEP8_CMD_TUPLE = (sys.executable,
+                          os.path.join(ROOT_DIR, 'autopep8.py'),)
+
 
 def py27_and_above(func):
     if sys.version_info[0] == 2 and sys.version_info[1] < 7:
@@ -56,8 +66,7 @@ class TestFixPEP8Error(unittest.TestCase):
         f = open(self.tempfile[1], 'w')
         f.write(line)
         f.close()
-        root_dir = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
-        cmd = [os.path.join(root_dir, 'autopep8.py')]
+        cmd = list(AUTOPEP8_CMD_TUPLE)
         cmd.extend(options.split())
         cmd.append(self.tempfile[1])
         p = Popen(cmd, stdout=PIPE)
@@ -304,6 +313,20 @@ class Foo():
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e225_with_indentation_fix(self):
+        line = """
+class Foo(object):
+  def bar(self):
+    return self.elephant!=None
+""".lstrip()
+        fixed = """
+class Foo(object):
+    def bar(self):
+        return self.elephant != None
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e231(self):
         line = "[1,2,3]\n"
         fixed = "[1, 2, 3]\n"
@@ -340,6 +363,12 @@ class Foo():
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e251_with_escaped_newline(self):
+        line = "1\n\n\ndef a(arg=\\\n1):\n    print(arg)\n"
+        fixed = "1\n\n\ndef a(arg=1):\n    print(arg)\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e261(self):
         line = "print 'a b '# comment\n"
         fixed = "print 'a b '  # comment\n"
@@ -352,12 +381,24 @@ class Foo():
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e261_with_dictionary_no_space(self):
+        line = "d = {#comment\n1: 2}\n"
+        fixed = "d = {  # comment\n1: 2}\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e262_more_space(self):
         line = "print 'a b '  #  comment\n"
         fixed = "print 'a b '  # comment\n"
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e262_with_no_comment(self):
+        line = "1  #\n123\n"
+        fixed = "1\n123\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e262_none_space(self):
         line = "print 'a b '  #comment\n"
         fixed = "print 'a b '  # comment\n"
@@ -370,12 +411,24 @@ class Foo():
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e262_hash_in_string_and_multiple_hashes(self):
+        line = "print 'a b  #string'  #comment #comment\n"
+        fixed = "print 'a b  #string'  # comment #comment\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e262_empty_comment(self):
         line = "print 'a b'  #\n"
         fixed = "print 'a b'\n"
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e262_more_complex(self):
+        line = "print 'a b '  #comment\n123\n"
+        fixed = "print 'a b '  # comment\n123\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e301(self):
         line = "class k:\n    s = 0\n    def f():\n        print 1\n"
         fixed = "class k:\n    s = 0\n\n    def f():\n        print 1\n"
@@ -430,10 +483,22 @@ class Foo():
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e702_with_semicolon_in_string(self):
+        line = 'print(";");\n'
+        fixed = 'print(";")\n'
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e702_with_semicolon_in_string_to_the_right(self):
+        line = 'x = "x"; y = "y;y"\n'
+        fixed = 'x = "x"\ny = "y;y"\n'
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e702_skip_with_triple_quote(self):
         # We do not support this yet.
         # We would expect '"""\n      hello\n   """\n1\n'.
-        line = '"""\n      hello\n   """;1\n'
+        line = '"""\n      hello\n   """; 1\n'
         fixed = line
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
@@ -451,9 +516,7 @@ class TestFixPEP8Warning(unittest.TestCase):
         f = open(self.tempfile[1], 'w')
         f.write(line)
         f.close()
-        root_dir = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
-        p = Popen([os.path.join(root_dir, 'autopep8.py'),
-                   self.tempfile[1]], stdout=PIPE)
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1]], stdout=PIPE)
         self.result = p.communicate()[0].decode('utf8')
 
     def test_w291(self):
@@ -486,6 +549,12 @@ class TestFixPEP8Warning(unittest.TestCase):
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_w391_more_complex(self):
+        line = "123\n456\n  \n"
+        fixed = "123\n456\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_w601(self):
         line = "a = {0: 1}\na.has_key(0)\n"
         fixed = "a = {0: 1}\n0 in a\n"
@@ -540,6 +609,16 @@ class TestFixPEP8Warning(unittest.TestCase):
         self._inner_setup(line)
         self.assertEqual(self.result, line)
 
+    def test_w601_skip_multi_line(self):
+        # We don't support this case
+        line = """
+a.has_key(
+0)
+""".lstrip()
+        fixed = line
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     @only_py2
     def test_w602_arg_is_string(self):
         line = "raise ValueError, \"w602 test\"\n"
@@ -560,6 +639,12 @@ class TestFixPEP8Warning(unittest.TestCase):
         self._inner_setup(line)
         self.assertEqual(self.result, line)
 
+    @only_py2
+    def test_w602_skip_logic(self):
+        line = "raise TypeError, e or 'hello'\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, line)
+
     @only_py2
     def test_w602_triple_quotes(self):
         line = 'raise ValueError, """hello"""\n1\n'
@@ -732,9 +817,7 @@ class TestOptions(unittest.TestCase):
         f = open(self.tempfile[1], 'w')
         f.write(line)
         f.close()
-        root_dir = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
-        p = Popen([os.path.join(root_dir, 'autopep8.py'),
-                   self.tempfile[1]] + options,
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1]] + options,
                   stdout=PIPE)
         self.result = p.communicate()[0].decode('utf8')
 
@@ -751,9 +834,8 @@ class TestOptions(unittest.TestCase):
         self.assertEqual(self.result, fixed)
 
     @py27_and_above
-    def test_no_arguments(self):
-        root_dir = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
-        p = Popen([os.path.join(root_dir, 'autopep8.py')],
+    def test_help(self):
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) + ['-h'],
                   stdout=PIPE)
         self.assertIn('Usage:', p.communicate()[0].decode('utf8'))
 
@@ -763,9 +845,7 @@ class TestOptions(unittest.TestCase):
         f = open(self.tempfile[1], 'w')
         f.write(line)
         f.close()
-        root_dir = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
-        p = Popen([os.path.join(root_dir, 'autopep8.py'),
-                   self.tempfile[1], '--verbose'],
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1], '--verbose'],
                   stdout=PIPE, stderr=PIPE)
         verbose_error = p.communicate()[1].decode('utf8')
         self.assertIn("'fix_e501' is not defined", verbose_error)
@@ -778,14 +858,61 @@ class TestOptions(unittest.TestCase):
         f = open(self.tempfile[1], 'w')
         f.write(line)
         f.close()
-        root_dir = os.path.split(os.path.abspath(os.path.dirname(__file__)))[0]
-        p = Popen([os.path.join(root_dir, 'autopep8.py'),
-                   self.tempfile[1], '--in-place'])
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1], '--in-place'])
         p.wait()
 
         f = open(self.tempfile[1])
         self.assertEqual(f.read(), fixed)
         f.close()
 
+
+class TestSpawnPEP8Process(unittest.TestCase):
+
+    def setUp(self):
+        self.tempfile = mkstemp()
+
+    def tearDown(self):
+        os.remove(self.tempfile[1])
+
+    def _inner_setup(self, line, options=""):
+        f = open(self.tempfile[1], 'w')
+        f.write(line)
+        f.close()
+        cmd = list(AUTOPEP8_CMD_TUPLE)
+        cmd.extend(options.split())
+        cmd.append(self.tempfile[1])
+
+        # Put fake pep8 path in front. It will cause autopep8 to launch a
+        # a subprocess by pretending to be an older version of pep8.
+        tmp_env = os.environ.copy()
+
+        if 'PYTHONPATH' in os.environ:
+            old_python_path = os.environ['PYTHONPATH']
+        else:
+            old_python_path = ''
+
+        tmp_env['PYTHONPATH'] = (os.path.join(ROOT_DIR, 'test',
+                                              'fake_pep8', 'site-packages') +
+                                 ':' + old_python_path)
+
+        if 'PATH' in os.environ:
+            old_path = os.environ['PATH']
+        else:
+            old_path = ''
+
+        tmp_env['PATH'] = (os.path.join(ROOT_DIR, 'test',
+                                        'fake_pep8', 'bin') +
+                           ':' + old_path)
+
+        p = Popen(cmd, stdout=PIPE, env=tmp_env)
+        self.result = p.communicate()[0].decode('utf8')
+
+    def test_basic(self):
+        line = "print('abc')    \n"
+        fixed = "print('abc')\n"
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+
 if __name__ == '__main__':
     unittest.main()
